@@ -1,3 +1,7 @@
+# Copyright (c) 2017 pandas-gbq Authors All rights reserved.
+# Use of this source code is governed by a BSD-style
+# license that can be found in the LICENSE file.
+
 import logging
 import time
 import warnings
@@ -14,102 +18,50 @@ except ImportError:  # pragma: NO COVER
     bigquery = None
     google_exceptions = None
 
-from pandas_gbq.exceptions import AccessDenied
-from pandas_gbq.exceptions import PerformanceWarning
+from pandas_gbq.exceptions import (
+    AccessDenied,
+    GenericGBQException,
+    PerformanceWarning,
+)
+from pandas_gbq import features
+from pandas_gbq.features import FEATURES
 import pandas_gbq.schema
 import pandas_gbq.timestamp
 
 
 logger = logging.getLogger(__name__)
 
-BIGQUERY_INSTALLED_VERSION = None
-BIGQUERY_CLIENT_INFO_VERSION = "1.12.0"
-BIGQUERY_BQSTORAGE_VERSION = "1.24.0"
-HAS_CLIENT_INFO = False
-HAS_BQSTORAGE_SUPPORT = False
-
 try:
     import tqdm  # noqa
 except ImportError:
     tqdm = None
 
 
-def _check_google_client_version():
-    global BIGQUERY_INSTALLED_VERSION, HAS_CLIENT_INFO, HAS_BQSTORAGE_SUPPORT, SHOW_VERBOSE_DEPRECATION
-
-    try:
-        import pkg_resources
-
-    except ImportError:
-        raise ImportError("Could not import pkg_resources (setuptools).")
-
-    # https://github.com/googleapis/python-bigquery/blob/master/CHANGELOG.md
-    bigquery_minimum_version = pkg_resources.parse_version("1.11.0")
-    bigquery_client_info_version = pkg_resources.parse_version(
-        BIGQUERY_CLIENT_INFO_VERSION
-    )
-    bigquery_bqstorage_version = pkg_resources.parse_version(
-        BIGQUERY_BQSTORAGE_VERSION
-    )
-    BIGQUERY_INSTALLED_VERSION = pkg_resources.get_distribution(
-        "google-cloud-bigquery"
-    ).parsed_version
-
-    HAS_CLIENT_INFO = (
-        BIGQUERY_INSTALLED_VERSION >= bigquery_client_info_version
-    )
-    HAS_BQSTORAGE_SUPPORT = (
-        BIGQUERY_INSTALLED_VERSION >= bigquery_bqstorage_version
-    )
-
-    if BIGQUERY_INSTALLED_VERSION < bigquery_minimum_version:
-        raise ImportError(
-            "pandas-gbq requires google-cloud-bigquery >= {0}, "
-            "current version {1}".format(
-                bigquery_minimum_version, BIGQUERY_INSTALLED_VERSION
-            )
-        )
-
-    # Add check for Pandas version before showing deprecation warning.
-    # https://github.com/pydata/pandas-gbq/issues/157
-    pandas_installed_version = pkg_resources.get_distribution(
-        "pandas"
-    ).parsed_version
-    pandas_version_wo_verbosity = pkg_resources.parse_version("0.23.0")
-    SHOW_VERBOSE_DEPRECATION = (
-        pandas_installed_version >= pandas_version_wo_verbosity
-    )
-
-
 def _test_google_api_imports():
+    try:
+        import pkg_resources  # noqa
+    except ImportError as ex:
+        raise ImportError("pandas-gbq requires setuptools") from ex
 
     try:
         import pydata_google_auth  # noqa
     except ImportError as ex:
-        raise ImportError(
-            "pandas-gbq requires pydata-google-auth: {0}".format(ex)
-        )
+        raise ImportError("pandas-gbq requires pydata-google-auth") from ex
 
     try:
         from google_auth_oauthlib.flow import InstalledAppFlow  # noqa
     except ImportError as ex:
-        raise ImportError(
-            "pandas-gbq requires google-auth-oauthlib: {0}".format(ex)
-        )
+        raise ImportError("pandas-gbq requires google-auth-oauthlib") from ex
 
     try:
         import google.auth  # noqa
     except ImportError as ex:
-        raise ImportError("pandas-gbq requires google-auth: {0}".format(ex))
+        raise ImportError("pandas-gbq requires google-auth") from ex
 
     try:
         from google.cloud import bigquery  # noqa
     except ImportError as ex:
-        raise ImportError(
-            "pandas-gbq requires google-cloud-bigquery: {0}".format(ex)
-        )
-
-    _check_google_client_version()
+        raise ImportError("pandas-gbq requires google-cloud-bigquery") from ex
 
 
 class DatasetCreationError(ValueError):
@@ -120,14 +72,6 @@ class DatasetCreationError(ValueError):
     pass
 
 
-class GenericGBQException(ValueError):
-    """
-    Raised when an unrecognized Google API Error occurs.
-    """
-
-    pass
-
-
 class InvalidColumnOrder(ValueError):
     """
     Raised when the provided column order for output
@@ -360,9 +304,7 @@ class GbqConnector(object):
             self.project_id = default_project
 
         if self.project_id is None:
-            raise ValueError(
-                "Could not determine project ID and one was not supplied."
-            )
+            raise ValueError("Could not determine project ID and one was not supplied.")
 
         # Cache the credentials if they haven't been set yet.
         if context.credentials is None:
@@ -416,16 +358,14 @@ class GbqConnector(object):
         # In addition to new enough version of google-api-core, a new enough
         # version of google-cloud-bigquery is required to populate the
         # client_info.
-        if HAS_CLIENT_INFO:
+        if FEATURES.bigquery_has_client_info:
             return bigquery.Client(
                 project=self.project_id,
                 credentials=self.credentials,
                 client_info=client_info,
             )
 
-        return bigquery.Client(
-            project=self.project_id, credentials=self.credentials
-        )
+        return bigquery.Client(project=self.project_id, credentials=self.credentials)
 
     @staticmethod
     def process_http_error(ex):
@@ -434,9 +374,7 @@ class GbqConnector(object):
 
         raise GenericGBQException("Reason: {0}".format(ex))
 
-    def run_query(
-        self, query, max_results=None, progress_bar_type=None, **kwargs
-    ):
+    def run_query(self, query, max_results=None, progress_bar_type=None, **kwargs):
         from concurrent.futures import TimeoutError
         from google.auth.exceptions import RefreshError
 
@@ -474,9 +412,7 @@ class GbqConnector(object):
             logger.debug("Query running...")
         except (RefreshError, ValueError):
             if self.private_key:
-                raise AccessDenied(
-                    "The service account credentials are not valid"
-                )
+                raise AccessDenied("The service account credentials are not valid")
             else:
                 raise AccessDenied(
                     "The credentials have been revoked or expired, "
@@ -491,9 +427,9 @@ class GbqConnector(object):
         while query_reply.state != "DONE":
             self.log_elapsed_seconds("  Elapsed", "s. Waiting...")
 
-            timeout_ms = job_config.get("jobTimeoutMs") or job_config[
-                "query"
-            ].get("timeoutMs")
+            timeout_ms = job_config.get("jobTimeoutMs") or job_config["query"].get(
+                "timeoutMs"
+            )
             timeout_ms = int(timeout_ms) if timeout_ms else None
             if timeout_ms and timeout_ms < self.get_elapsed_seconds() * 1000:
                 raise QueryTimeout("Query timeout: {} ms".format(timeout_ms))
@@ -518,8 +454,7 @@ class GbqConnector(object):
             bytes_billed = query_reply.total_bytes_billed or 0
             logger.debug(
                 "Query done.\nProcessed: {} Billed: {}".format(
-                    self.sizeof_fmt(bytes_processed),
-                    self.sizeof_fmt(bytes_billed),
+                    self.sizeof_fmt(bytes_processed), self.sizeof_fmt(bytes_billed),
                 )
             )
             logger.debug(
@@ -537,11 +472,7 @@ class GbqConnector(object):
         )
 
     def _download_results(
-        self,
-        query_job,
-        max_results=None,
-        progress_bar_type=None,
-        user_dtypes=None,
+        self, query_job, max_results=None, progress_bar_type=None, user_dtypes=None,
     ):
         # No results are desired, so don't bother downloading anything.
         if max_results == 0:
@@ -550,14 +481,15 @@ class GbqConnector(object):
         if user_dtypes is None:
             user_dtypes = {}
 
-        if self.use_bqstorage_api and not HAS_BQSTORAGE_SUPPORT:
+        if self.use_bqstorage_api and not FEATURES.bigquery_has_bqstorage:
             warnings.warn(
                 (
                     "use_bqstorage_api was set, but have google-cloud-bigquery "
                     "version {}. Requires google-cloud-bigquery version "
                     "{} or later."
                 ).format(
-                    BIGQUERY_INSTALLED_VERSION, BIGQUERY_BQSTORAGE_VERSION
+                    FEATURES.bigquery_installed_version,
+                    features.BIGQUERY_BQSTORAGE_VERSION,
                 ),
                 PerformanceWarning,
                 stacklevel=4,
@@ -568,18 +500,14 @@ class GbqConnector(object):
             create_bqstorage_client = False
 
         to_dataframe_kwargs = {}
-        if HAS_BQSTORAGE_SUPPORT:
-            to_dataframe_kwargs[
-                "create_bqstorage_client"
-            ] = create_bqstorage_client
+        if FEATURES.bigquery_has_bqstorage:
+            to_dataframe_kwargs["create_bqstorage_client"] = create_bqstorage_client
 
         try:
             query_job.result()
             # Get the table schema, so that we can list rows.
             destination = self.client.get_table(query_job.destination)
-            rows_iter = self.client.list_rows(
-                destination, max_results=max_results
-            )
+            rows_iter = self.client.list_rows(destination, max_results=max_results)
 
             schema_fields = [field.to_api_repr() for field in rows_iter.schema]
             conversion_dtypes = _bqschema_to_nullsafe_dtypes(schema_fields)
@@ -587,7 +515,7 @@ class GbqConnector(object):
             df = rows_iter.to_dataframe(
                 dtypes=conversion_dtypes,
                 progress_bar_type=progress_bar_type,
-                **to_dataframe_kwargs
+                **to_dataframe_kwargs,
             )
         except self.http_error as ex:
             self.process_http_error(ex)
@@ -604,11 +532,11 @@ class GbqConnector(object):
     def load_data(
         self,
         dataframe,
-        dataset_id,
-        table_id,
+        destination_table_ref,
         chunksize=None,
         schema=None,
         progress_bar=True,
+        api_method: str = "load_parquet",
     ):
         from pandas_gbq import load
 
@@ -618,11 +546,11 @@ class GbqConnector(object):
             chunks = load.load_chunks(
                 self.client,
                 dataframe,
-                dataset_id,
-                table_id,
+                destination_table_ref,
                 chunksize=chunksize,
                 schema=schema,
                 location=self.location,
+                api_method=api_method,
             )
             if progress_bar and tqdm:
                 chunks = tqdm.tqdm(chunks)
@@ -635,74 +563,8 @@ class GbqConnector(object):
         except self.http_error as ex:
             self.process_http_error(ex)
 
-    def schema(self, dataset_id, table_id):
-        """Retrieve the schema of the table
-
-        Obtain from BigQuery the field names and field types
-        for the table defined by the parameters
-
-        Parameters
-        ----------
-        dataset_id : str
-            Name of the BigQuery dataset for the table
-        table_id : str
-            Name of the BigQuery table
-
-        Returns
-        -------
-        list of dicts
-            Fields representing the schema
-        """
-        table_ref = self.client.dataset(dataset_id).table(table_id)
-
-        try:
-            table = self.client.get_table(table_ref)
-            remote_schema = table.schema
-
-            remote_fields = [
-                field_remote.to_api_repr() for field_remote in remote_schema
-            ]
-            for field in remote_fields:
-                field["type"] = field["type"].upper()
-                field["mode"] = field["mode"].upper()
-
-            return remote_fields
-        except self.http_error as ex:
-            self.process_http_error(ex)
-
-    def verify_schema(self, dataset_id, table_id, schema):
-        """Indicate whether schemas match exactly
-
-        Compare the BigQuery table identified in the parameters with
-        the schema passed in and indicate whether all fields in the former
-        are present in the latter. Order is not considered.
-
-        Parameters
-        ----------
-        dataset_id :str
-            Name of the BigQuery dataset for the table
-        table_id : str
-            Name of the BigQuery table
-        schema : list(dict)
-            Schema for comparison. Each item should have
-            a 'name' and a 'type'
-
-        Returns
-        -------
-        bool
-            Whether the schemas match
-        """
-
-        fields_remote = pandas_gbq.schema._clean_schema_fields(
-            self.schema(dataset_id, table_id)
-        )
-        fields_local = pandas_gbq.schema._clean_schema_fields(schema["fields"])
-        return fields_remote == fields_local
-
     def delete_and_recreate_table(self, dataset_id, table_id, table_schema):
-        table = _Table(
-            self.project_id, dataset_id, credentials=self.credentials
-        )
+        table = _Table(self.project_id, dataset_id, credentials=self.credentials)
         table.delete(table_id)
         table.create(table_id, table_schema)
 
@@ -760,9 +622,7 @@ def _cast_empty_df_dtypes(schema_fields, df):
     ``object``.
     """
     if not df.empty:
-        raise ValueError(
-            "DataFrame must be empty in order to cast non-nullsafe dtypes"
-        )
+        raise ValueError("DataFrame must be empty in order to cast non-nullsafe dtypes")
 
     dtype_map = {"BOOLEAN": bool, "INTEGER": np.int64}
 
@@ -946,7 +806,7 @@ def read_gbq(
 
     _test_google_api_imports()
 
-    if verbose is not None and SHOW_VERBOSE_DEPRECATION:
+    if verbose is not None and FEATURES.pandas_has_deprecated_verbose:
         warnings.warn(
             "verbose is deprecated and will be removed in "
             "a future version. Set logging level in order to vary "
@@ -983,9 +843,7 @@ def read_gbq(
             final_df.set_index(index_col, inplace=True)
         else:
             raise InvalidIndexColumn(
-                'Index column "{0}" does not exist in DataFrame.'.format(
-                    index_col
-                )
+                'Index column "{0}" does not exist in DataFrame.'.format(index_col)
             )
 
     # Change the order of columns in the DataFrame based on provided list
@@ -993,9 +851,7 @@ def read_gbq(
         if sorted(col_order) == sorted(final_df.columns):
             final_df = final_df[col_order]
         else:
-            raise InvalidColumnOrder(
-                "Column order does not match this DataFrame."
-            )
+            raise InvalidColumnOrder("Column order does not match this DataFrame.")
 
     connector.log_elapsed_seconds(
         "Total time taken",
@@ -1017,6 +873,7 @@ def to_gbq(
     location=None,
     progress_bar=True,
     credentials=None,
+    api_method: str = "default",
     verbose=None,
     private_key=None,
 ):
@@ -1037,7 +894,8 @@ def to_gbq(
     dataframe : pandas.DataFrame
         DataFrame to be written to a Google BigQuery table.
     destination_table : str
-        Name of table to be written, in the form ``dataset.tablename``.
+        Name of table to be written, in the form ``dataset.tablename`` or
+        ``project.dataset.tablename``.
     project_id : str, optional
         Google BigQuery Account project ID. Optional when available from
         the environment.
@@ -1069,11 +927,12 @@ def to_gbq(
     table_schema : list of dicts, optional
         List of BigQuery table fields to which according DataFrame
         columns conform to, e.g. ``[{'name': 'col1', 'type':
-        'STRING'},...]``.
+        'STRING'},...]``.  The ``type`` values must be BigQuery type names.
 
         - If ``table_schema`` is provided, it may contain all or a subset of
           DataFrame columns. If a subset is provided, the rest will be
-          inferred from the DataFrame dtypes.
+          inferred from the DataFrame dtypes.  If ``table_schema`` contains
+          columns not in the DataFrame, they'll be ignored.
         - If ``table_schema`` is **not** provided, it will be
           generated according to dtypes of DataFrame columns. See
           `Inferring the Table Schema
@@ -1104,6 +963,12 @@ def to_gbq(
         :class:`google.oauth2.service_account.Credentials` directly.
 
         .. versionadded:: 0.8.0
+    api_method : str, optional
+        API method used to upload DataFrame to BigQuery. One of "load_parquet",
+        "load_csv". Default "load_parquet" if pandas is version 1.1.0+,
+        otherwise "load_csv".
+
+        .. versionadded:: 0.16.0
     verbose : bool, deprecated
         Deprecated in Pandas-GBQ 0.4.0. Use the `logging module
         to adjust verbosity instead
@@ -1119,7 +984,7 @@ def to_gbq(
 
     _test_google_api_imports()
 
-    if verbose is not None and SHOW_VERBOSE_DEPRECATION:
+    if verbose is not None and FEATURES.pandas_has_deprecated_verbose:
         warnings.warn(
             "verbose is deprecated and will be removed in "
             "a future version. Set logging level in order to vary "
@@ -1128,12 +993,35 @@ def to_gbq(
             stacklevel=1,
         )
 
+    if api_method == "default":
+        # Avoid using parquet if pandas doesn't support lossless conversions to
+        # parquet timestamp. See: https://stackoverflow.com/a/69758676/101923
+        if FEATURES.pandas_has_parquet_with_lossless_timestamp:
+            api_method = "load_parquet"
+        else:
+            api_method = "load_csv"
+
+    if chunksize is not None:
+        if api_method == "load_parquet":
+            warnings.warn(
+                "chunksize is ignored when using api_method='load_parquet'",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        elif api_method == "load_csv":
+            warnings.warn(
+                "chunksize will be ignored when using api_method='load_csv' in a future version of pandas-gbq",
+                PendingDeprecationWarning,
+                stacklevel=2,
+            )
+
     if if_exists not in ("fail", "replace", "append"):
         raise ValueError("'{0}' is not valid for if_exists".format(if_exists))
 
     if "." not in destination_table:
         raise NotFoundException(
-            "Invalid Table Name. Should be of the form 'datasetId.tableId' "
+            "Invalid Table Name. Should be of the form 'datasetId.tableId' or "
+            "'projectId.datasetId.tableId'"
         )
 
     connector = GbqConnector(
@@ -1145,7 +1033,14 @@ def to_gbq(
         private_key=private_key,
     )
     bqclient = connector.client
-    dataset_id, table_id = destination_table.rsplit(".", 1)
+
+    destination_table_ref = bigquery.table.TableReference.from_string(
+        destination_table, default_project=connector.project_id
+    )
+
+    project_id_table = destination_table_ref.project
+    dataset_id = destination_table_ref.dataset_id
+    table_id = destination_table_ref.table_id
 
     default_schema = _generate_bq_schema(dataframe)
     if not table_schema:
@@ -1157,10 +1052,10 @@ def to_gbq(
 
     # If table exists, check if_exists parameter
     try:
-        table = bqclient.get_table(destination_table)
+        table = bqclient.get_table(destination_table_ref)
     except google_exceptions.NotFound:
         table_connector = _Table(
-            project_id,
+            project_id_table,
             dataset_id,
             location=location,
             credentials=connector.credentials,
@@ -1177,21 +1072,17 @@ def to_gbq(
                 "'append' or 'replace' data."
             )
         elif if_exists == "replace":
-            connector.delete_and_recreate_table(
-                dataset_id, table_id, table_schema
-            )
+            connector.delete_and_recreate_table(dataset_id, table_id, table_schema)
         elif if_exists == "append":
-            if not pandas_gbq.schema.schema_is_subset(
-                original_schema, table_schema
-            ):
+            if not pandas_gbq.schema.schema_is_subset(original_schema, table_schema):
                 raise InvalidSchema(
                     "Please verify that the structure and "
                     "data types in the DataFrame match the "
                     "schema of the destination table."
                 )
 
-            # Update the local `table_schema` so mode matches.
-            # See: https://github.com/pydata/pandas-gbq/issues/315
+            # Update the local `table_schema` so mode (NULLABLE/REQUIRED)
+            # matches. See: https://github.com/pydata/pandas-gbq/issues/315
             table_schema = pandas_gbq.schema.update_schema(
                 table_schema, original_schema
             )
@@ -1203,11 +1094,11 @@ def to_gbq(
 
     connector.load_data(
         dataframe,
-        dataset_id,
-        table_id,
+        destination_table_ref,
         chunksize=chunksize,
         schema=table_schema,
         progress_bar=progress_bar,
+        api_method=api_method,
     )
 
 
@@ -1224,8 +1115,7 @@ def generate_bq_schema(df, default_type="STRING"):
     """
     # deprecation TimeSeries, #11121
     warnings.warn(
-        "generate_bq_schema is deprecated and will be removed in "
-        "a future version",
+        "generate_bq_schema is deprecated and will be removed in " "a future version",
         FutureWarning,
         stacklevel=2,
     )
@@ -1265,6 +1155,15 @@ class _Table(GbqConnector):
             private_key=private_key,
         )
 
+    def _table_ref(self, table_id):
+        """Return a BigQuery client library table reference"""
+        from google.cloud.bigquery import DatasetReference
+        from google.cloud.bigquery import TableReference
+
+        return TableReference(
+            DatasetReference(self.project_id, self.dataset_id), table_id
+        )
+
     def exists(self, table_id):
         """Check if a table exists in Google BigQuery
 
@@ -1280,7 +1179,7 @@ class _Table(GbqConnector):
         """
         from google.api_core.exceptions import NotFound
 
-        table_ref = self.client.dataset(self.dataset_id).table(table_id)
+        table_ref = self._table_ref(table_id)
         try:
             self.client.get_table(table_ref)
             return True
@@ -1300,31 +1199,25 @@ class _Table(GbqConnector):
             Use the generate_bq_schema to generate your table schema from a
             dataframe.
         """
-        from google.cloud.bigquery import SchemaField
+        from google.cloud.bigquery import DatasetReference
         from google.cloud.bigquery import Table
+        from google.cloud.bigquery import TableReference
 
         if self.exists(table_id):
-            raise TableCreationError(
-                "Table {0} already " "exists".format(table_id)
-            )
+            raise TableCreationError("Table {0} already exists".format(table_id))
 
         if not _Dataset(self.project_id, credentials=self.credentials).exists(
             self.dataset_id
         ):
             _Dataset(
-                self.project_id,
-                credentials=self.credentials,
-                location=self.location,
+                self.project_id, credentials=self.credentials, location=self.location,
             ).create(self.dataset_id)
 
-        table_ref = self.client.dataset(self.dataset_id).table(table_id)
+        table_ref = TableReference(
+            DatasetReference(self.project_id, self.dataset_id), table_id
+        )
         table = Table(table_ref)
-
-        schema = pandas_gbq.schema.add_default_nullable_mode(schema)
-
-        table.schema = [
-            SchemaField.from_api_repr(field) for field in schema["fields"]
-        ]
+        table.schema = pandas_gbq.schema.to_google_cloud_bigquery(schema)
 
         try:
             self.client.create_table(table)
@@ -1344,7 +1237,7 @@ class _Table(GbqConnector):
         if not self.exists(table_id):
             raise NotFoundException("Table does not exist")
 
-        table_ref = self.client.dataset(self.dataset_id).table(table_id)
+        table_ref = self._table_ref(table_id)
         try:
             self.client.delete_table(table_ref)
         except NotFound:
@@ -1371,6 +1264,12 @@ class _Dataset(GbqConnector):
             private_key=private_key,
         )
 
+    def _dataset_ref(self, dataset_id):
+        """Return a BigQuery client library dataset reference"""
+        from google.cloud.bigquery import DatasetReference
+
+        return DatasetReference(self.project_id, dataset_id)
+
     def exists(self, dataset_id):
         """Check if a dataset exists in Google BigQuery
 
@@ -1387,7 +1286,7 @@ class _Dataset(GbqConnector):
         from google.api_core.exceptions import NotFound
 
         try:
-            self.client.get_dataset(self.client.dataset(dataset_id))
+            self.client.get_dataset(self._dataset_ref(dataset_id))
             return True
         except NotFound:
             return False
@@ -1409,7 +1308,7 @@ class _Dataset(GbqConnector):
                 "Dataset {0} already " "exists".format(dataset_id)
             )
 
-        dataset = Dataset(self.client.dataset(dataset_id))
+        dataset = Dataset(self._dataset_ref(dataset_id))
 
         if self.location is not None:
             dataset.location = self.location
